import numpy as np

from pymoo.core.problem import Problem
from pymoo.problems.meta import MetaProblem


class Surrogate:

    def __init__(self,
                 problem,
                 targets=None,
                 **kwargs):

        """

        This surrogate object allows to conveniently build and update surrogates for a Population object.
        This can be a rather complicated task because surrogates for objectives and ieq. and eq. constraints might
        need to be build and different combinations of doing that are possible.

        Parameters
        ----------
        problem : Problem
            The optimization problem to access the meta data (it will never be called for an evaluation)

        targets : list
            A list of target objects describing how each of the components of a population should be modeled.
            This also includes the type of model and other hyper-parameters. This modular definition is necessary
            because different types of surrogate might be used for different _targets.

        """

        super().__init__(**kwargs)
        self._problem = problem
        self.targets = targets if targets is not None else []

    def validate(self, trn=None, tst=None, **kwargs):
        for target in self.targets:
            target.validate(trn=trn, tst=tst, **kwargs)

    def fit(self, sols):
        for target in self.targets:
            target.fit(sols)

    def performance(self, indicator, **kwargs):
        ret = {}
        for target in self.targets:
            ret[target.label] = target.performance(indicator, **kwargs)
        return ret

    def problem(self):
        return ProblemFromTargets(self._problem, self.targets)


class ProblemFromTargets(MetaProblem):

    def __init__(self, problem, targets, **kwargs):
        super().__init__(problem, **kwargs)
        self.targets = targets

    def _evaluate(self, X, out, *args, **kwargs):
        n = len(X)

        out["F"] = np.full((n, self.n_obj), np.nan, dtype=float)
        out["G"] = np.full((n, self.n_ieq_constr), np.nan, dtype=float)
        out["H"] = np.full((n, self.n_eq_constr), np.nan, dtype=float)

        for target in self.targets:
            target.predict(X, out)

        for v in ["F", "G", "H"]:
            if np.any(np.isnan(out[v])):
                raise Exception("Building Surrogate has failed (nan values were predict). The run has been terminated.")

        out["F_estm_error"] = np.full((n, self.n_obj), np.nan, dtype=float)
        out["G_estm_error"] = np.full((n, self.n_ieq_constr), np.nan, dtype=float)
        out["H_estm_error"] = np.full((n, self.n_eq_constr), np.nan, dtype=float)

        for target in self.targets:
            label, k = target.label
            out.get(label + "_estm_error")[:, k] = target.performance("mae")



import numpy as np

from pymoo.core.sampling import Sampling
from pymoo.util.normalization import normalize, denormalize
from pymoo.util.ref_dirs.energy import squared_dist, calc_potential_energy_with_grad
from pymoo.util.ref_dirs.optimizer import Adam
from pysamoo.sampling.niching import NichingConstrainedSampling
from pysamoo.sampling.rejection import RejectionConstrainedSampling


def calc_potential_energy(A, d):
    i, j = np.triu_indices(len(A), 1)
    D = np.sqrt(squared_dist(A, A)[i, j])
    energy = np.log((1 / D ** d).mean())
    return energy


class EnergyConstrainedSampling(Sampling):

    def __init__(self,
                 func_eval_constr,
                 n_max_iter=10000):
        super().__init__()
        self.func_eval_constr = func_eval_constr
        self.n_max_iter = n_max_iter

    def _do(self, problem, n_samples, **kwargs):
        xl, xu = problem.bounds()
        constr = self.func_eval_constr
        d = problem.n_var ** 2

        X = RejectionConstrainedSampling(constr).do(problem, n_samples).get("X")
        if len(X) < n_samples:
            X = NichingConstrainedSampling(constr).do(problem, n_samples).get("X")

        if len(X) == 0:
            raise Exception("No feasible solution could be found!")
        elif len(X) < n_samples:
            print("WARNING: Less feasible solutions than requested could be found.")

        X = normalize(X, xl, xu)

        optimizer = Adam(alpha=0.005)

        obj, grad = calc_potential_energy_with_grad(X, d)
        hist = [obj]

        done = False

        for i in range(self.n_max_iter):

            if done:
                break

            _X = optimizer.next(X, grad)
            _CV = constr(denormalize(_X, xl, xu))
            feasible = np.logical_and(_CV <= 0, np.all(np.logical_and(_X >= 0, _X <= 1), axis=1))

            X[feasible] = _X[feasible]

            obj, grad = calc_potential_energy_with_grad(X, d)
            hist.append(obj)

            hist = hist[-100:]

            avg_impr = (- np.diff(hist[-100:])).mean()

            if len(hist) > 100:
                if avg_impr < 1e-3:
                    optimizer = Adam(alpha=optimizer.alpha / 2)
                elif avg_impr < 1e-6:
                    done = True

        X = denormalize(X, xl, xu)

        return X



from ezmodel.core.factory import models_from_clazzes
from ezmodel.models.kriging import Kriging
from ezmodel.models.rbf import RBF
from ezmodel.util.transformation.plog import Plog
from pymoo.util.normalization import NoNormalization


def DEFAULT_OBJ_MODELS(**defaults):

    models = models_from_clazzes(
        # YAGP,
        # BOTORCH,
        # TKGP,
        # Kriging,
        # GGP,
        # LGP,
        RBF,
        # KNN,
        # RBF3,
        # PolynomialRegression,
        # RBF2,
        # pySOTRBF,
        # SVR,
        # InverseDistanceWeighting,
        # NearestNeighbors,
        **defaults)

    models = {name: entry["model"] for name, entry in models.items()}

    # models = {}

    # for kernel in ["cubic", "linear", "mq"]:
    #     for normalized in [False, True]:
    #         for tail in ["constant", "linear", "linear+quadratic"]:
    #             params = dict(defaults)
    #             params["kernel"] = kernel
    #             params["normalized"] = normalized
    #             params["tail"] = tail
    #
    #             model = RBF(**params)
    #             models[f"rbf-{kernel}-{tail}-{normalized}"] = model

    models['kriging-const'] = Kriging(regr="constant")
    models['kriging-lin'] = Kriging(regr="linear")
    models['kriging-quadr'] = Kriging(regr="quadratic")

    models['kriging-const-ARD'] = Kriging(regr="constant", ARD=True)
    models['kriging-lin-ARD'] = Kriging(regr="linear", ARD=True)
    models['kriging-quadr-ARD'] = Kriging(regr="quadratic", ARD=True)
    # models['kriging-sine'] = Kriging(regr="sine")

    return models


def DEFAULT_IEQ_CONSTR_MODELS(**defaults):
    models = {}

    for kernel in ["cubic", "linear", "mq"]:
        for label, norm in [("default", NoNormalization()), ("plog", Plog())]:
            for normalized in [False, True]:
                for tail in ["constant", "linear", "linear+quadratic"]:
                    params = dict(defaults)
                    params["kernel"] = kernel
                    params["norm_X"] = norm
                    params["normalized"] = normalized
                    params["tail"] = tail

                    model = RBF(**params)
                    models[f"rbf-{kernel}-{tail}-{label}-{normalized}"] = model

    return models


def DEFAULT_EQ_CONSTR_MODELS(**defaults):
    models = {}

    for kernel in ["cubic", "linear", "mq"]:
        for normalized in [False, True]:
            for tail in ["constant", "linear", "linear+quadratic"]:
                for optimize in [False]:
                    params = dict(defaults)
                    params["kernel"] = kernel
                    params["normalized"] = normalized
                    params["tail"] = tail
                    params["optimize"] = optimize

                    model = RBF(**params)
                    models[f"rbf-{kernel}-{tail}-{normalized}-{optimize}"] = model

    models['kriging-const'] = Kriging(regr="constant")
    models['kriging-lin'] = Kriging(regr="linear")
    models['kriging-quadr'] = Kriging(regr="quadratic")

    return models



from copy import deepcopy

import numpy as np

from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.evaluator import Evaluator
from pymoo.core.population import Population
from pymoo.util.display import Display
from pymoo.util.dominator import get_relation
from pymoo.util.misc import norm_eucl_dist
from pymoo.util.normalization import ZeroToOneNormalization
from pymoo.util.termination.no_termination import NoTermination
from pysamoo.core.algorithm import SurrogateAssistedAlgorithm
from pysamoo.core.surrogate import Surrogate
from pysamoo.core.target import Target
from ezmodel.core.factory import models_from_clazzes
from ezmodel.models.knn import KNN
from ezmodel.models.kriging import Kriging
from ezmodel.models.rbf import RBF


class PSAFDisplay(Display):

    def __init__(self, display, **kwargs):
        super().__init__(**kwargs)
        self.display = display

    def _do(self, problem, evaluator, psaf):
        self.display.do(problem, evaluator, psaf.algorithm, show=False)
        self.output = self.display.output
        self.output.append("bias", "-" if psaf.bias is None else psaf.bias)

        if problem.n_obj == 1 and problem.n_constr == 0:
            target = psaf.surrogate.targets[0]
            self.output.append("mae", target.performance("mae"))
            self.output.append("model", target.best)


class PSAF(SurrogateAssistedAlgorithm):

    def __init__(self, algorithm, alpha=5, beta=30, rho=None, rho_max=0.7, eps=0.005, **kwargs):
        SurrogateAssistedAlgorithm.__init__(self, **kwargs)
        self.algorithm = deepcopy(algorithm)
        self.alpha = alpha
        self.beta = beta
        self.eps = eps
        self.rho = rho
        self.rho_max = rho_max
        self.bias = rho

    def _setup(self, problem, **kwargs):
        assert problem.n_obj == 1 and problem.n_constr == 0, "PSAF only works for unconstrained single-objective problems!"

        super()._setup(problem, **kwargs)

        self.algorithm.setup(problem, **kwargs)

        xl, xu = problem.bounds()
        defaults = dict(norm_X=ZeroToOneNormalization(xl, xu))

        models = models_from_clazzes(RBF, **defaults)
        models = {name: entry["model"] for name, entry in models.items()}
        models = {**models, **{"krg-cont": Kriging(regr="constant"), "krg-lin": Kriging(regr="linear")}}

        if "baseline" not in models:
            models["baseline"] = KNN(problem.n_var + 1)

        # create the ensemble of surrogates considering objectives and constraints
        targets = [Target(("F", 0), models)]
        self.surrogate = Surrogate(problem, targets)

        # customize the display to show the surrogate influence
        self.display = PSAFDisplay(self.algorithm.display)

        # define the survival for the individuals to keep
        self.survival = FitnessSurvival() if self.problem.n_obj == 1 else RankAndCrowdingSurvival()

    def _initialize_advance(self, infills=None, **kwargs):

        # validate and check different surrogates to find the best
        self.surrogate.validate(infills, exclude=["baseline"])

        # now we perform a fake initialization of the algorithm object by providing individuals from LHS
        fake = self.algorithm.infill()
        fittest = self.survival.do(self.problem, infills, n_survive=len(fake))
        self.algorithm.advance(infills=fittest)

        super()._initialize_advance(infills=infills, **kwargs)

    def _infill(self):
        problem, algorithm, surrogate = self.surrogate.problem(), self.algorithm, self.surrogate

        # advance the surrogate which results in doing a benchmark and setting the best surrogate combination
        surrogate.fit(self.archive)

        # set the bias the surrogate is supposed to have - only if greater than zero we do the second phase
        target = surrogate.targets[0]
        r2 = 1 - (target.performance("mse") / target.performance("mse", model="baseline"))
        bias = max(self.rho_max, r2) if self.rho is None else self.rho

        # calculate the infill solutions as the algorithms usually would
        off = algorithm.infill()

        # use the ensemble to values the current offsprings
        Evaluator().eval(problem, off)

        # if a tournament selection should be done alpha is at least two
        if self.alpha > 1:

            # do the tournament for each alpha
            for k in range(self.alpha - 1):

                # create a second pool and actually do the tournament
                others = self.algorithm.infill()
                Evaluator().eval(problem, others)

                # for each offspring see if we do the surrogate tournament
                for k in range(len(off)):

                    # if the competitor is not worse it will take the lead
                    if get_relation(others[k], off[k]) >= 0:
                        off[k] = others[k]

        # if algorithm shall be continued on the surrogate and there is a bias at all
        if self.beta > 0 and bias > 0.0:

            # already calculate what individuals will be replaced later
            replace = np.random.random(len(off)) <= bias

            # if at least one is replaced actually simulate the algorithm on the surrogate
            if replace.sum() > 0:

                # create a copy of the algorithm object
                algorithm = deepcopy(self.algorithm)
                algorithm.termination = NoTermination()
                Evaluator().eval(problem, algorithm.pop, skip_already_evaluated=False)

                # the candidates to take from beta are initialized to the offsprings
                cands = off.copy()

                # run the algorithm for beta generations and always assign replacement candidates
                for k in range(self.beta):

                    # just to make sure no algorithm specific termination criterion has be executed
                    if not algorithm.has_next():
                        break

                    # get the infills from the algorithm and evaluate on the surrogate
                    infills = algorithm.infill()
                    Evaluator().eval(problem, infills)

                    # if there is some candidates to consider
                    if len(infills) > 0:

                        # find the closest individuals for each candidate to offsprings
                        dists = norm_eucl_dist(problem, infills.get("X"), off.get("X"))
                        I = dists.argmin(axis=1)

                        # for each infill solution check if it replaces the candidate
                        for j in range(len(infills)):

                            # get the index to the closest of offsprings
                            i = I[j]

                            # if it is better indeed then we update it
                            if get_relation(infills[j], cands[i]) >= 0:
                                cands[i] = infills[j]

                    # now continue the execution of the algorithm
                    algorithm.advance(infills=infills)

                # now do the probabilistic replacement
                for i in range(len(off)):

                    # if it should be replaced (that has been pre-calculated)
                    if replace[i]:
                        off[i] = cands[i]

        self.bias = bias

        # no only use the X values
        infills = Population.new(X=off.get("X"))

        # set the velocity for PSO
        infills.set("V", off.get("V"), "index", off.get("index"))

        return infills

    def _advance(self, infills=None, **kwargs):
        # update the surrogate(s) with the new infills points
        self.surrogate.validate(self.archive, infills, exclude=["baseline"])

        # make a step in the main algorithm with high-fidelity solutions
        self.algorithm.advance(infills=infills, **kwargs)

        super()._advance(infills=infills, **kwargs)

    def _set_optimum(self):
        self.opt = self.algorithm.opt





import numpy as np

try:
    from smac.facade.func_facade import fmin_smac
except:
    raise Exception("smac not found. Please execute: 'pip install smac'")



from pymoo.algorithms.base.local import LocalSearch
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.operators.sampling.lhs import LatinHypercubeSampling
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.optimum import filter_optimum


class FunctionCall:

    def __init__(self, problem) -> None:
        super().__init__()
        self.problem = problem

    def __call__(self, x):
        return self.problem.evaluate(x[None, :])[0, 0]


class SMAC(LocalSearch):

    def __init__(self, x0=None,
                 sampling=LatinHypercubeSampling(),
                 display=SingleObjectiveDisplay(),
                 n_sample_points="auto",
                 n_max_sample_points=50,
                 **kwargs):

        super().__init__(x0, sampling, display, n_sample_points, n_max_sample_points, **kwargs)

        self.cnt = 0
        self.history = None

    def _local_advance(self, **kwargs):

        if self.history is None:
            self.history = self._optimize()

        self.evaluator.n_eval += 1
        self.pop = self.history.pop(0)
        self.cnt += 1

        if len(self.history) == 0:
            self.termination.force_termination = True

    def _optimize(self):

        problem = self.problem

        func = FunctionCall(problem)

        seed = np.random.randint(0, 2 ** 32 - 1)

        eval_remaininig = self.termination.n_max_evals - self.evaluator.n_eval

        x, cost, obj = fmin_smac(func=func,
                                 x0=self.x0.X,
                                 bounds=np.column_stack(problem.bounds()),
                                 maxfun=eval_remaininig,
                                 rng=seed,
                                 scenario_args=dict(output_dir=None))

        pop = []

        for k, v in obj.runhistory.data.items():
            config_id = k.config_id
            f = v.cost
            x = np.array(list(obj.runhistory.ids_config[config_id]._values.values()))
            ind = Individual(X=x, F=np.array([f]), CV=np.array([0.0]), feasible=[True])
            pop.append(Population.create(ind))

        return pop

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)





import random
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.duplicate import DefaultDuplicateElimination
from pymoo.core.evaluator import Evaluator
from pymoo.core.population import Population
from pymoo.operators.sampling.lhs import LHS
from pymoo.operators.selection.tournament import compare
from pymoo.optimize import default_termination
from pymoo.util.display import Display
from pymoo.util.dominator import get_relation
from pymoo.util.misc import norm_eucl_dist, cdist
from pymoo.util.optimum import filter_optimum
from pymoo.util.termination.no_termination import NoTermination
from pymoo.visualization.fitness_landscape import FitnessLandscape
from pymoo.visualization.video.callback_video import AnimationCallback

from pysamoo.core.algorithm import SurrogateAssistedAlgorithm
# =========================================================================================================
# Display
# =========================================================================================================
from pysamoo.core.knockout import noisy


class GPSAFDisplay(Display):

    def __init__(self, display, **kwargs):
        super().__init__(**kwargs)
        self.display = display

    def _do(self, problem, evaluator, gpasf):
        self.display.do(problem, evaluator, gpasf.algorithm, show=False)
        self.output = self.display.output

        if gpasf.n_gen > 1:
            surr_infills = Population.create(*gpasf.infills.get("created_by"))
            n_influenced = sum(surr_infills.get("type") == "trace")
            self.output.append("n_influenced", f"{n_influenced}/{len(surr_infills)}")
        else:
            self.output.append("n_influenced", "-")

        if problem.n_obj == 1 and problem.n_constr == 0:
            if len(gpasf.surrogate.targets) >= 1:
                target = gpasf.surrogate.targets[0]
                self.output.append("mae", target.performance("mae"))
                self.output.append("model", target.best)

        elif problem.n_obj == 2 and problem.n_constr == 0:
            if len(gpasf.surrogate.targets) >= 2:
                perf = gpasf.surrogate.performance("mae")
                if ("F", 0) in perf:
                    self.output.append("mae f1", perf[("F", 0)])
                if ("F", 1) in perf:
                    self.output.append("mae f2", perf[("F", 1)])


# =========================================================================================================
# Animation
# =========================================================================================================

class GPSAFAnimation(AnimationCallback):

    def __init__(self,
                 nth_gen=1,
                 n_samples_for_surface=200,
                 dpi=200,
                 **kwargs):

        super().__init__(nth_gen=nth_gen, dpi=dpi, **kwargs)
        self.n_samples_for_surface = n_samples_for_surface
        self.last_pop = None

    def do(self, problem, algorithm):

        if problem.n_var != 2 or problem.n_obj != 1:
            raise Exception(
                "This visualization can only be used for problems with two variables and one objective!")

        # draw the problem surface
        doe = algorithm.surrogate.targets["F"].doe
        if doe is not None:
            problem = algorithm.surrogate

        plot = FitnessLandscape(problem, _type="contour", kwargs_contour=dict(alpha=0.5))
        plot.do()

        if doe is not None:
            plt.scatter(doe.get("X")[:, 0], doe.get("X")[:, 1], color="black", alpha=0.3)

        for k, sols in enumerate(algorithm.trace_assigned):
            if len(sols) > 0:
                pop = Population.create(*sols)
                plt.scatter(pop.get("X")[:, 0], pop.get("X")[:, 1], color="blue", alpha=0.3)

                x = algorithm.influenced[k].X
                for sol in sols:
                    plt.plot((x[0], sol.X[0]), (x[1], sol.X[1]), alpha=0.1, color="black")

        plt.scatter(algorithm.influenced.get("X")[:, 0], algorithm.influenced.get("X")[:, 1], color="red", marker="*",
                    alpha=0.7,
                    label="influenced")

        _biased = Population.create(
            *[e for e in algorithm.biased if e is not None])
        plt.scatter(_biased.get("X")[:, 0], _biased.get("X")[:, 1], color="orange", marker="s", label="Selected",
                    alpha=0.8,
                    s=100)

        plt.legend()


# =========================================================================================================
# Algorithm
# =========================================================================================================


class GPSAF(SurrogateAssistedAlgorithm):

    def __init__(self,
                 algorithm,
                 alpha=5,
                 beta=30,
                 rho=None,
                 n_max_doe=100,
                 n_max_infills=np.inf,
                 **kwargs):

        SurrogateAssistedAlgorithm.__init__(self, **kwargs)

        self.proto = deepcopy(algorithm)
        self.algorithm = None

        # the control parameters for the surrogate assistance
        self.alpha = alpha
        self.beta = beta
        self.rho = rho
        self.n_max_doe = n_max_doe

        # the maximum number of infill solutions
        self.n_max_infills = n_max_infills

        self.surr_infills = None
        self.influenced, self.trace, self.biased = None, None, None
        self.restart = False

    def _setup(self, problem, **kwargs):
        super()._setup(problem, **kwargs)

        # set the default termination to the proto type
        if self.proto.termination is None:
            self.proto.termination = default_termination(problem)

        # setup the underlying algorithm
        self.proto.setup(problem, seed=self.seed, **kwargs)

        # copy the algorithm object to get started
        self.algorithm = deepcopy(self.proto)

        # customize the display to show the surrogate influence
        self.display = GPSAFDisplay(self.algorithm.display)

        # define the survival for the individuals to keep
        self.survival = FitnessSurvival() if problem.n_obj == 1 else RankAndCrowdingSurvival()

    def _initialize_infill(self):
        # return self.algorithm.infill()
        return super()._initialize_infill()

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills=infills, **kwargs)

        # validate and check different surrogates to find the best
        self.surrogate.validate(infills)

        # now we perform a fake initialization of the algorithm object by providing individuals from LHS
        # fake = self.algorithm.infill()
        # fittest = self.survival.do(self.problem, infills, n_survive=len(fake))
        fittest = infills

        # feed back the fittest individuals to the algorithm
        self.algorithm.advance(infills=fittest)

    def _infill(self):

        # if the algorithm should do a restart - copy again the prototype
        if self.restart:
            self.algorithm = deepcopy(self.proto)

        # get the design of experiments to be used for modeling
        doe = self._doe()

        # always before using the surrogate fit it
        self.surrogate.fit(doe)

        # get the error estimate from the surrogate
        error = self.surrogate.performance("mae")

        # this would be the default behavior of the algorithm
        influenced = self.algorithm.infill()

        # do the alpha phase based on the tournament selection based on the surrogate prediction
        if self.alpha is not None:
            influenced = self._infill_alpha(influenced, error=error)

        # now by default the infills are the surrogate-influenced solutions
        infills = influenced
        infills.set("type", "influenced")

        # continue running the algorithm for more generations if beta is "enabled"
        if self.beta > 0:

            # get the trace from the beta run on the surrogate
            trace = self._infill_beta()
            trace.set("type", "trace")
            self.trace = trace

            # 3) assign the found solutions to the original infill solutions
            trace_assigned = self._infill_beta_assign(influenced, trace)

            # when a solution is considered a duplicate
            eps = 1e-12

            # for each solution from the surrogate based optimization
            for i, pool in enumerate(trace_assigned):

                # if there are no solutions to replace continue directly
                if len(pool) == 0:
                    continue

                # create a population object from the pool
                pool = Population.create(*pool)

                # the probability of replacement
                rho = self.rho

                # if not fixed we use the distribution of points
                if rho is None:
                    rho = (len(pool) / max([len(e) for e in trace_assigned])) ** 0.5

                # if the solution should be replaced
                if np.random.random() <= rho:

                    # if it should be replaced find the ONE solution from the pool
                    biased = self._infill_prob_tourn(pool, method="tournament", error=error, n_winners=1)[0]

                    # check the distance to existing solutions
                    closest = norm_eucl_dist(self.problem, biased.get("X"), self.archive.get("X")).min()

                    # if the solution is in fact new
                    if closest > eps:

                        # now actually set the value to the infills
                        infills[i] = biased
                        # infills[i].X = biased.X

                    else:
                        print("BIASED: TOO CLOSE (SKIP)")

                closest = norm_eucl_dist(self.problem, infills[i].get("X"), self.archive.get("X")).min()

                # if the solution is in fact new
                if closest < eps:
                    print("INFLUENCED: TOO CLOSE")

        # if beta is zero, then simply take the results from the alpha phase
        else:
            infills = influenced

        # now copy over the infills and set them to have never been evaluated
        ret = infills.copy(deep=True)
        for e in ret:
            e.reset(data=False)
        ret.set("X", infills.get("X"), "created_by", infills)

        return ret

    def _infill_alpha(self, infills, error=None):
        problem = self.surrogate.problem()

        # create the influenced population using the tournament
        influenced = infills

        # use the ensemble to values the current offsprings
        Evaluator().eval(problem, influenced)

        # reduce the number of infill solutions if required
        if len(influenced) > self.n_max_infills:
            influenced = self._infill_prob_tourn(influenced, method="tournament", error=error, n_winners=self.n_max_infills)

        # do the tournament for each alpha
        for k in range(self.alpha - 1):

            # create a second pool and actually do the tournament
            others = self.algorithm.infill()
            if len(others) > len(influenced):
                I = np.random.permutation(len(others))[:len(influenced)]
                others = others[I]

            Evaluator().eval(problem, others)

            # for each offspring see if we do the surrogate tournament
            for k in range(len(influenced)):

                # if the competitor is not worse it will take the lead
                if get_relation(influenced[k], others[k]) < 1:
                    influenced[k] = others[k]

        return influenced

    def _infill_beta_assign(self, influenced, trace, filter=False):
        ret = [[] for _ in range(len(influenced))]

        if len(trace) > 0:

            # find the closest individuals for each candidate to offsprings (and NOT the current replaced one)
            dists = norm_eucl_dist(self.problem, trace.get("X"), influenced.get("X"))
            closest = dists.argmin(axis=1)

            for k, i in enumerate(closest):

                # add the solution closest to the influence to the list
                if not filter or get_relation(influenced[i], trace[k]) < 1:
                    trace[k].set("i", i)
                    ret[i].append(trace[k])

        return ret

    def _infill_beta(self):

        # define the problem on the surrogate
        problem = self.surrogate.problem()

        # create a copy of the algorithm object to keep the original unmodified
        algorithm = deepcopy(self.algorithm)
        # if np.random.random() < 0.5:
        #     algorithm = deepcopy(self.algorithm)
        # else:
        #     algorithm = deepcopy(self.proto)
        #     print("PROTO")

        # disable the termination to have have enough iterations to continue
        algorithm.termination = NoTermination()
        algorithm.has_terminated = False

        # now use the surrogate to evaluate the solutions from the current algorithm object
        if algorithm.pop is not None:
            Evaluator().eval(problem, algorithm.pop, skip_already_evaluated=False)

        trace = []

        # run the algorithm for beta iterations
        for k in range(self.beta):

            # just to make sure no algorithm specific termination criterion has be executed
            if not algorithm.has_next():
                break

            # get the infills from the algorithm and evaluate on the surrogate
            infills = algorithm.infill()
            infills.set("k", k)
            Evaluator().eval(problem, infills)

            trace.extend(infills)

            # now continue the execution of the algorithm
            algorithm.advance(infills=infills)

        return Population.create(*trace)

    def _infill_prob_tourn(self, sols, n_winners=1, method="tournament", error=None):

        # if the beta phase has not found any solutions close to the influenced one
        if len(sols) == 0:
            return None

        else:

            if method == "best":
                return FitnessSurvival().do(self.problem, Population.create(*sols), n_survive=n_winners)[0]

            elif method == "random":
                return np.random.choice(sols, size=n_winners)

            elif method == "tournament":

                # create a copy of all solutions to be considered
                pool = list(range(len(sols)))

                # until we have found a clear winner of the tournament
                while True:

                    # always shuffle the pool to have random tournaments
                    random.shuffle(pool)

                    # make sure the pool is an even number
                    if len(pool) % 2 != 0:
                        pool.append(np.random.choice(pool))

                    # create the pairs that compete with each other
                    pairs = np.reshape(np.array(pool), (-1, 2))

                    # prepare the next pool already containing all the winners
                    winners = []

                    # create a solution pool with noise
                    sols_with_noise = noisy(sols, error)

                    for i, j in pairs:

                        # the two solutions to be compared
                        a, b = sols_with_noise[i], sols_with_noise[j]

                        # calc the relation in a probabilistic manner
                        rel = get_relation(a, b)

                        if rel == 1:
                            winners.append(i)
                        elif rel == -1:
                            winners.append(j)
                        else:
                            if a.get("k") is not None and b.get("k") is not None:
                                k = compare(i, a.get("k"), j, b.get("k"), method='larger_is_better',
                                            return_random_if_equal=True)
                            else:
                                k = np.random.choice([i, j])

                            winners.append(k)

                    if len(winners) <= n_winners:

                        if len(winners) < n_winners:
                            H = set(winners)

                            random.shuffle(pool)

                            for k in pool:

                                if k not in H:
                                    winners.append(k)
                                    H.add(k)

                                if len(winners) == n_winners:
                                    break

                        return sols[winners]

                    pool = winners

            else:
                raise Exception("Unknown selection.")

    def _advance(self, infills=None, **kwargs):

        if self.restart:

            for k in np.random.permutation(len(infills)):

                opt = np.random.choice(self.opt)
                if get_relation(opt, infills[k]) >= 0:
                    infills[k] = opt
                    break

            self.restart = False

        # validate the current model
        self.surrogate.validate(trn=self.doe, tst=infills)

        # make a step in the main algorithm with high-fidelity solutions
        self.algorithm.advance(infills=infills, **kwargs)

        if not self.algorithm.has_next():
            # self.restart = True
            # print("RESTART")

            self.restart = False
            print("RESTART is DISABLED")

        # for target in self.surrogate.targets:
        #     print(target.label, target.best)


        super()._advance(infills=infills, **kwargs)

    def _doe(self):
        n_max_doe = self.n_max_doe

        doe = self.archive

        if len(doe) > n_max_doe:

            center = LHS().do(self.problem, n_max_doe).get("X")
            A = cdist(doe.get("X"), center).argmin(axis=1)

            cluster = [[] for _ in range(n_max_doe)]
            for i, a in enumerate(A):
                cluster[a].append(i)

            doe = []
            for c in cluster:
                if len(c) > 0:
                    s = np.random.choice(c)
                    doe.append(s)

            doe = self.archive[doe]
            doe = Population.merge(doe, self.algorithm.pop)
            doe = Population.merge(doe, self.algorithm.opt)
            doe = Population.merge(doe, self.opt)
            doe = DefaultDuplicateElimination().do(doe)

            # if we have constraints also select the closest infeas. or feas. solution
            others = []
            if self.problem.has_constraints():

                is_feas = self.archive.get("feas")
                feas, infeas = np.where(is_feas)[0], np.where(~is_feas)[0]

                for sol in doe:
                    if sol.get("feas"):
                        if len(infeas) > 0:
                            s = infeas[cdist(sol.get("X")[None, :], self.archive[infeas].get("X"))[0].argmin()]
                            others.append(s)
                    else:
                        if len(feas) > 0:
                            s = feas[cdist(sol.get("X")[None, :], self.archive[feas].get("X"))[0].argmin()]
                            others.append(s)

            doe = Population.merge(doe, self.archive[others])
            doe = DefaultDuplicateElimination().do(doe)

        self.doe = doe

        return doe

    def _set_optimum(self):
        sols = self.algorithm.opt
        if self.opt is not None:
            sols = Population.merge(sols, self.opt)
        self.opt = filter_optimum(sols, least_infeasible=True)



import numpy as np
from scipy.stats import rankdata


def calc_mse(y_true, y_hat, **kwargs):
    return ((y_true - y_hat) ** 2).mean()


def calc_rmse(y_true, y_hat, **kwargs):
    return calc_mse(y_true, y_hat, **kwargs) ** 0.5


def calc_mae(y_true, y_hat, **kwargs):
    return np.abs(y_true - y_hat).mean()


def kendall_tau(y_true, y_hat, trn_y=None, **kwargs):
    assert trn_y is not None, "For kendall tau the ranking needs to be calculated which requires the training data!"

    a = rankdata(np.concatenate([trn_y, y_true]), method='min')
    b = rankdata(np.concatenate([trn_y, y_hat]), method='min')

    n = len(a)

    i, j = np.meshgrid(np.arange(n), np.arange(n))

    ndisordered = np.logical_or(np.logical_and(a[i] < a[j], b[i] > b[j]),
                                np.logical_and(a[i] > a[j], b[i] < b[j])).sum()

    # ndisordered = ndisordered / (n * (n - 1))

    return ndisordered


def calc_r2(y_true, y_hat, trn_y=None, **kwargs):
    if trn_y is None:
        trn_y = y_true

    return 1 - (calc_mse(y_true, y_hat) / calc_mse(trn_y, trn_y.mean()))


def calc_sign_error(y_true, y_hat, **kwargs):
    return (np.sign(y_true) != np.sign(y_hat)).mean()


INDICATORS = {
    "rmse": dict(sign=+1, func=calc_rmse),
    "mse": dict(sign=+1, func=calc_mse),
    "mae": dict(sign=+1, func=calc_mae),
    "kendall_tau": dict(sign=+1, func=kendall_tau),
    "r2": dict(sign=-1, func=calc_r2),
    "sign": dict(sign=-1, func=calc_sign_error),
}



import numpy as np
import scipy

from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.constraints.tcv import TotalConstraintViolation
from pymoo.core.algorithm import Algorithm
from pymoo.core.population import Population
from pymoo.operators.sampling.lhs import LHS
from pymoo.operators.sampling.rnd import random
from pymoo.optimize import minimize
from pymoo.problems.meta import MetaProblem
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.misc import cdist
from pymoo.util.normalization import NoNormalization
from pymoo.util.optimum import filter_optimum
from pysamoo.core.surrogate import Surrogate
from ezmodel.models.rbf import RBF
from ezmodel.util.transformation.plog import Plog
from ezmodel.util.transformation.zero_to_one import ZeroToOneNormalization


# =========================================================================================================
# Display
# =========================================================================================================


class SACOBRADisplay(SingleObjectiveDisplay):

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        self.output.append("eps", algorithm.eps)
        self.output.append("rho", algorithm.rho if algorithm.rho is not None else "-")
        self.output.append("C_feas", int(algorithm.C_feas))
        self.output.append("C_infeas", int(algorithm.C_infeas))


class SACOBRA(Algorithm):

    def __init__(self, n_doe=None, display=SACOBRADisplay(), **kwargs):
        super().__init__(display=display, **kwargs)
        self.n_doe = n_doe
        self.archive = Population()

        self.rho = None

        self.C_feas = 0
        self.C_infeas = 0

        self.T_feas = None
        self.T_infeas = None

        self.repair = None

    def _setup(self, problem, **kwargs):

        # directly rescale the whole problem
        problem = Rescale(problem)

        if self.n_doe is None:
            self.n_doe = 2 * problem.n_var + 1

        xl, xu = problem.bounds()
        l = (xu - xl).min()
        self.eps = 0.005 * l
        self.eps_max = 0.01 * l

        self.T_feas = np.floor(2 * np.sqrt(problem.n_var))
        self.T_infeas = self.T_feas

        #
        models = {}

        for label, norm in [("default", NoNormalization()), ("plog", Plog())]:
            model = RBF(kernel='cubic', tail='linear', normalized=False, norm_y=norm)
            models[f"rbf-{label}"] = model

        targets = {
            "F": ObjectivesAsTarget({k: v for (k, v) in models.items() if "plog" not in k}),
            "G": ConstraintsAsTarget(models),
        }

        self.surrogate = Surrogate(problem, targets)

        self.problem = problem

    def _initialize_infill(self):
        return LHS().do(self.problem, self.n_doe)

    def _initialize_advance(self, infills=None, **kwargs):
        self.archive = Population.merge(self.archive, infills)

        # analyze the initial population and find ranges
        self._analyze_initial_pop(infills)

        # adjust the constraint values
        self._adjust_constr(infills)

        # adjust the DRC parameter based on whether the obj. function is steep or not
        self._adjust_drc()

        # initialize the surrogate with the doe points
        self.surrogate.initialize(infills)

    def _infill(self):

        # update the current surrogate(s) based on the most recent update done after the last generation
        self.surrogate.advance()

        # and now finally create the models
        self.surrogate.fit(metric="spear", minimize=False)

        verbose = False

        if verbose:

            for target, entry in self.surrogate.targets["F"].predictor.predictors.items():
                print(target, entry["label"], entry["mae"])

            for target, entry in self.surrogate.targets["G"].predictor.predictors.items():
                print(target, entry["label"], entry["mae"])

        # get the current best solution and the infill
        feas = self.archive.get("feas")
        opt, sol = self.opt[0], self.archive[-1]

        # if the new solution has a better objective value than the current one, flag the next iteration for repair
        if not np.any(feas):

            x_new = self._repair()

        elif not sol.feas and sol.f < opt.f:

            # attempt to repair the promising solution
            x_new = self._repair(sol)

        # this is the regular iteration when obj. and constr. are used for optimization
        else:

            # find the rho by cycling through the XI array
            self.rho = self.XI[(self.n_gen - 1) % len(self.XI)]

            x_start = self._random_restart()

            # search on the surrogate (cobra iteration)
            x_new = self._cobra(x_start, self.eps, self.rho)

        return Population.new(X=[x_new])

    def _advance(self, infills=None, **kwargs):

        # adjust the constraint values
        self._adjust_constr(infills)

        # adjust the eps value (margin for infeasibility)
        self._adjust_margins(infills)

        # add them to the archive
        self.archive = Population.merge(self.archive, infills)

        # add the new infill solutions to the surrogate to be used in the next iteration
        self.surrogate.update(infills)

    def _analyze_initial_pop(self, pop):
        f, G = pop.get("f", "G")
        self.FR = f.max() - f.min()
        self.GR = G.max(axis=0) - G.min(axis=0)

    def _adjust_constr(self, pop):
        G = pop.get("G")
        _G = G * (self.GR.mean() / self.GR)
        pop.set("G", _G)

    def _adjust_margins(self, infills):
        sol = infills[0]

        if sol.feas:
            self.C_feas += 1
            self.C_infeas = 0
        else:
            self.C_feas = 0
            self.C_infeas += 1

        if self.C_feas >= self.T_feas:
            self.eps = 0.5 * self.eps
            self.C_feas = 0

        if self.C_infeas >= self.T_infeas:
            self.eps = min(2 * self.eps, self.eps_max)
            self.C_infeas = 0

    def _cobra(self, x_start, eps, rho):

        X = self.archive.get("X")

        tcv = TotalConstraintViolation(ieq_eps=-1 * eps)

        def func_obj(x):
            return self.surrogate.evaluate(x, return_values_of=["F"])[0]

        def func_constr(x):
            G = self.surrogate.evaluate(x, return_values_of=["G"])
            cv = tcv.calc(G)
            return -1 * (cv[0])

        def func_constr_trust(x):
            closest = cdist(x[None, :], X).min()
            trust = rho - closest
            return -1 * trust

        constraints = [{"type": 'ineq', "fun": func_constr}, {"type": 'ineq', "fun": func_constr_trust}]

        xl, xu = self.problem.bounds()
        bounds = np.column_stack([xl, xu])

        options = {}

        res = scipy.optimize.minimize(func_obj,
                                      x_start,
                                      args=(),
                                      bounds=bounds,
                                      method='SLSQP',
                                      constraints=constraints,
                                      tol=None,
                                      callback=None,
                                      options=options)

        print(func_obj(res.x), func_constr(res.x), func_constr_trust(res.x))
        print(res.x)
        return res.x

    def _random_restart(self):
        p1 = 0.125
        p2 = 0.4

        feas = self.archive.get("feas")

        if feas.sum() / len(feas) < 0.05:
            p = p2
        else:
            p = p1

        if np.random.random() < p:
            x = random(self.problem)[0]
        else:
            x = self.opt[0].X

        return x

    def _repair(self, sol=None):

        tcv = TotalConstraintViolation()

        class MyProblem(MetaProblem):

            def _evaluate(self, x, out, *args, **kwargs):
                super()._evaluate(x, out, *args, **kwargs)

                if sol is not None:
                    out["F"] = cdist(x, sol.X[None, :])
                else:
                    out["F"] = tcv.calc(out["G"])
                    out["G"][:] = 0.0

        problem = MyProblem(self.surrogate)

        sampling = Population.new(X=self.archive.get("X"))
        self.evaluator.eval(problem, sampling, count_evals=False)

        # let us start with a maximum of 100 solutions
        if len(sampling) > 100:
            sampling = FitnessSurvival().do(self.surrogate, sampling, n_survive=100)

        algorithm = DE(pop_size=len(sampling), sampling=sampling)

        res = minimize(problem,
                       algorithm,
                       return_least_infeasible=True)

        return res.X

    def _adjust_drc(self):
        if self.FR > 1000:
            self.XI = np.array([0.001, 1e-16])
        else:
            self.XI = np.array([0.3, 0.05, 0.001, 0.0005, 1e-16])

    def _set_optimum(self):
        self.opt = filter_optimum(self.archive, least_infeasible=True)


class MinusOneToOneNormalization(ZeroToOneNormalization):

    def forward(self, X):
        return super().forward(X) * 2 - 1

    def backward(self, X):
        return super().backward((X + 1) / 2)


class Rescale(MetaProblem):

    def __init__(self, problem):
        super().__init__(problem)
        assert self.xl is not None and self.xu is not None, "Both, xl and xu, must be set to redefine the problem!"

        self.norm = MinusOneToOneNormalization(problem.xl, problem.xu)
        self.xl, self.xu = -np.ones(self.n_var), np.ones(self.n_var)

    def do(self, x, out, *args, **kwargs):
        out["__X__"] = x
        xp = self.norm.backward(x)
        super().do(xp, out, *args, **kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        ps = super()._calc_pareto_set(*args, **kwargs)
        return self.norm.forward(ps)



from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.core.duplicate import DefaultDuplicateElimination
from pymoo.core.population import Population
from pymoo.optimize import minimize
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import normalize
from pymoo.util.roulette import RouletteWheelSelection
from sklearn.cluster import KMeans

from pysamoo.core.algorithm import SurrogateAssistedAlgorithm


class SSANSGA2(SurrogateAssistedAlgorithm):

    def __init__(self,
                 n_infills=10,
                 surr_pop_size=100,
                 surr_n_gen=30,
                 surr_eps_elim=1e-6,
                 surr_sampling="current",
                 display=MultiObjectiveDisplay(),
                 **kwargs):

        super().__init__(display=display, **kwargs)
        self.n_infills = n_infills
        self.surr_n_gen = surr_n_gen
        self.surr_pop_size = surr_pop_size
        self.surr_eps_elim = surr_eps_elim
        self.surr_sampling = surr_sampling

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.surrogate.validate(infills)

    def _infill(self):

        self.surrogate.fit(self.archive)

        problem = self.surrogate.problem()

        if self.surr_sampling == "current":
            sampling = self.archive
        elif self.surr_sampling == "random":
            sampling = None
        else:
            raise Exception("Unknown surrogate sampling strategy.")

        algorithm = NSGA2(pop_size=self.surr_pop_size,
                          sampling=sampling
                          )

        res = minimize(problem,
                       algorithm,
                       ('n_gen', self.surr_n_gen),
                       seed=1,
                       verbose=False)

        cand = DefaultDuplicateElimination(epsilon=self.surr_eps_elim).do(res.pop, self.archive)

        if len(cand) <= self.n_infills:
            infills = Population.new(X=cand.get("X"))

        else:

            ideal = res.opt.get("F").min(axis=0)
            nadir = res.opt.get("F").max(axis=0) + 1e-16
            vals = normalize(cand.get("F"), ideal, nadir)

            kmeans = KMeans(n_clusters=self.n_infills, random_state=0).fit(vals)
            groups = [[] for _ in range(self.n_infills)]
            for k, i in enumerate(kmeans.labels_):
                groups[i].append(k)

            S = []

            for group in groups:
                if len(group) > 0:
                    fitness = cand[group].get("crowding").argsort()
                    selection = RouletteWheelSelection(fitness, larger_is_better=False)
                    I = group[selection.next()]
                    S.append(I)

            infills = Population.new(X=cand[S].get("X"))

        return infills

    def _advance(self, infills=None, **kwargs):
        self.surrogate.validate(self.archive, infills)
        super()._advance(infills, **kwargs)

    def _set_optimum(self):
        nds = NonDominatedSorting().do(self.archive.get("F"), only_non_dominated_front=True)
        self.opt = self.archive[nds]



from pymoo.core.algorithm import Algorithm
from pymoo.core.initialization import Initialization
from pymoo.core.population import Population
from pymoo.operators.sampling.lhs import LHS
from pymoo.util.normalization import ZeroToOneNormalization

from pysamoo.core.defaults import DEFAULT_OBJ_MODELS, DEFAULT_IEQ_CONSTR_MODELS, DEFAULT_EQ_CONSTR_MODELS
from pysamoo.core.surrogate import Surrogate
from pysamoo.core.target import Target


def default_n_doe(n, max=float("inf")):
    return min(2 * n + 1, max)


class SurrogateAssistedAlgorithm(Algorithm):

    def __init__(self,
                 n_initial_doe=None,
                 n_initial_max_doe=100,
                 sampling=LHS(),
                 nth_validate=5,
                 surrogate=None,
                 **kwargs):
        """
        Parameters
        ----------
        n_initial_doe : int
            Number of initial design of experiments. If `None`, the default is 11*n - 1. (but at most `n_max_doe`)

        n_max_doe : int
            If `n_initial_doe` is set to `None`, the maximum number of initial designs.

        sampling : class
            The initial sampling being used for the designs of experiment.

        """
        super().__init__(**kwargs)

        self.n_initial_doe = n_initial_doe
        self.n_initial_max_doe = n_initial_max_doe
        self.initialization = Initialization(sampling)

        # all solutions that have been evaluated so far
        self.archive = Population()

        # here always the most recent infill solutions are stored
        self.infills = None

        # the model/surrogate to be used during optimization
        self.surrogate = surrogate

        # a solution set which has not been evaluated yet on the models
        self.validation = Population()

        # each nth iteration when all surrogate models should be revalidated
        self.nth_validate = nth_validate

    def _setup(self, problem, **kwargs):

        # initialize the default surrogate for the algorithm
        if self.surrogate is None:

            # the design space boundaries for the problem - used for normalization in the surrogate
            xl, xu = problem.bounds()
            defaults = dict(norm_X=MyNormalization(xl, xu))

            targets = []

            models = DEFAULT_OBJ_MODELS(**defaults)
            for m in range(problem.n_obj):
                target = Target(("F", m), models)
                targets.append(target)

            models = DEFAULT_IEQ_CONSTR_MODELS(**defaults)
            for g in range(problem.n_ieq_constr):
                target = Target(("G", g), models)
                targets.append(target)

            models = DEFAULT_EQ_CONSTR_MODELS(**defaults)
            for h in range(problem.n_eq_constr):
                target = Target(("H", h), models)
                targets.append(target)

            # create the surrogate model
            self.surrogate = Surrogate(problem, targets)

        # set the number of DOE points initially
        if self.n_initial_doe is None:
            self.n_initial_doe = min(self.n_initial_max_doe, default_n_doe(problem.n_var))

    def _initialize_infill(self):
        self.infills = self.initialization.do(self.problem, self.n_initial_doe, algorithm=self)
        return self.infills

    def _initialize_advance(self, infills=None, **kwargs):
        self.infills = infills
        self.archive = Population.merge(self.archive, infills)

    def _advance(self, infills=None, **kwargs):
        self.infills = infills
        self.archive = Population.merge(self.archive, infills)


class MyNormalization(ZeroToOneNormalization):

    def forward(self, X):
        return super().forward(X) * 200 - 100

    def backward(self, X):
        return super().backward((X + 100) / 200)



from scipy.stats import norm

# =========================================================================================================
# Acquisition Functions
# =========================================================================================================
from pymoo.problems.meta import MetaProblem


class AcquisitionFunction:

    def calc(self, mu, sigma, **kwargs):
        pass


class EI(AcquisitionFunction):

    def calc(self, mu, sigma, f_min=None, **kwargs):
        if f_min is None:
            raise Exception("Estimation of minimum function value needs to be provided!")

        f = - sigma

        # through precision error sigma can be negative - this should actually never be the case
        pos_sigma = sigma > 0
        mu, sigma = mu[pos_sigma], sigma[pos_sigma]

        # minimization version of EI
        impr = f_min - mu

        # calculate expected improvement
        z = impr / sigma

        ei = impr * norm.cdf(z) + sigma * norm.pdf(z)

        # because we are minimizing take the negative expected improvement
        f[pos_sigma] = - ei

        return f


class POI(AcquisitionFunction):

    def calc(self, mu, sigma, f_min=None, **kwargs):
        if f_min is None:
            raise Exception("Estimation of minimum function value needs to be provided!")

        pos_sigma = sigma > 0
        mu, sigma = mu[pos_sigma], sigma[pos_sigma]
        f = sigma

        # minimization version of PI
        impr = f_min - mu

        # calculate pi
        z = impr / sigma
        pi = norm.pdf(z)

        f[pos_sigma] = - pi

        return f


class UCB(AcquisitionFunction):

    def __init__(self, beta=3.0) -> None:
        super().__init__()
        self.beta = beta

    def calc(self, mu, sigma, **kwargs):
        ucb = mu - self.beta * sigma
        return ucb


# =========================================================================================================
# Acquisition Problem
# =========================================================================================================


class AcquisitionProblem(MetaProblem):

    def __init__(self,
                 problem,
                 model,
                 acquisition_func,
                 **kwargs):
        super().__init__(problem)
        self.model = model
        self.acquisition_func = acquisition_func
        self.kwargs = kwargs

    def _evaluate(self, x, out, *args, **kwargs):
        if self.model is None:
            raise Exception("Please set the model for the problem to be defined.")

        # calculate the metric using the implementation
        mu, sigma = self.model.predict(x, return_values_of=["y", "sigma"])
        mu, sigma = mu[:, 0], sigma[:, 0]

        # calculate the value of the acquisition function
        acq = self.acquisition_func.calc(mu, sigma, x=x, **self.kwargs)

        out["F"], out["acq"] = acq, acq



from collections import Counter

import numpy as np
from pymoo.constraints.tcv import TotalConstraintViolation
from pymoo.core.population import Population
from pymoo.core.replacement import ReplacementSurvival
from pymoo.util.dominator import get_relation


def is_better(a, b):
    assert len(a) == len(b)

    ret = np.full(len(a), False)

    a_f, a_cv, a_feas = a.get("f", "cv", "feas")
    b_f, b_cv, b_feas = b.get("f", "cv", "feas")

    # 1) Both infeasible and constraints have been improved
    ret[(~a_feas & ~b_feas) & (b_cv < a_cv)] = True

    # 2) A solution became feasible
    ret[~a_feas & b_feas] = True

    # 3) Both feasible but objective space value has improved
    ret[(a_feas & b_feas) & (b_f < a_f)] = True

    return ~ret


def comp(a, b, error=None):
    if error is not None:
        a = noisy(a, error)
        b = noisy(b, error)
    ret = is_better(a, b)
    return ret


def pcomp(sols, pairs, error=None):
    a, b = pairs.T
    a_is_better_than_b = comp(sols[a], sols[b], error=error)

    ret = np.copy(b)
    ret[a_is_better_than_b] = a[a_is_better_than_b]

    return ret


def knockout(sols, n_winners=1, error=None):
    # create a copy of all solutions to be considered
    pool = list(np.random.permutation(len(sols)))

    # until we have found a clear winner of the tournament
    while len(pool) > n_winners:

        # the list of winners in this round
        winners = []

        # make sure the pool is an even number, otherwise add one randomly (not equal to the other competing index)
        if len(pool) % 2 != 0:
            winners.append(pool[-1])
            pool = pool[:-1]

        # create the pairs that compete with each other
        pairs = np.reshape(pool, (-1, 2))

        # now add all the winners as well from this tournament
        W = pcomp(sols, pairs, error=error)
        winners.extend(W)

        # that means we have now less than we want - fill up with random solutions from pool
        if len(winners) < n_winners:
            S = set(winners)

            for k in np.random.permutation(len(sols)):

                # if not added yet then add it
                if k not in S:
                    winners.append(k)

                # we have filled up to the required number of winners -> done
                if len(winners) == n_winners:
                    break

        pool = winners

    return sols[pool]


class NoisyReplacement(ReplacementSurvival):

    def __init__(self, error, **kwargs):
        super().__init__(**kwargs)
        self.error = error

    def _do(self, problem, pop, off, **kwargs):
        return comp(off, pop, error=self.error)


def calc_prob_relation(a, b, error=None, n_comparisons=1):
    ret = []

    for k in range(n_comparisons):
        if error is not None:
            _a, _b = noisy(Population.create(a, b), error)
        else:
            _a, _b = a, b

        _rel = get_relation(_a, _b)
        ret.append(_rel)

    rel = np.random.choice([val for val, freq in Counter(ret).most_common()])

    return rel


def noisy(sols, error):
    out = {}
    for type in ["X", "F", "G", "H"]:
        out[type] = np.copy(sols.get(type))

    for (type, k), std in error.items():
        out[type][:, k] += np.random.normal(loc=0.0, scale=std, size=len(sols))

    noisy = Population.new(**out)
    TotalConstraintViolation().do(noisy, inplace=True)
    return noisy



from pymoo.algorithms.soo.nonconvex.ga_niching import NicheGA
from pymoo.core.sampling import Sampling
from pymoo.operators.sampling.lhs import LHS
from pymoo.optimize import minimize
from pymoo.problems.meta import MetaProblem
from pymoo.util.reference_direction import select_points_with_maximum_distance


class NichingConstrainedSampling(Sampling):

    def __init__(self, func_eval_constr, sampling=LHS(), initial_eps=0.25):
        super().__init__()
        self.func_eval_constr = func_eval_constr
        self.initial_eps = initial_eps
        self.sampling = sampling

    def _do(self, problem, n_samples, **kwargs):
        constr = self.func_eval_constr

        class ConstrainedProblem(MetaProblem):

            def __init__(self, problem):
                super().__init__(problem)
                self.n_obj = 1
                self.n_constr = 1

            def _evaluate(self, x, out, *args, **kwargs):
                cv = constr(x)
                out["F"] = cv
                out["G"] = cv

        problem = ConstrainedProblem(problem)

        eps = self.initial_eps

        while True:

            algorithm = NicheGA(pop_size=n_samples, samping=self.sampling, norm_niche_size=eps, norm_by_dim=True)

            res = minimize(problem, algorithm, ("n_gen", 200), return_least_infeasible=True)
            opt = res.opt

            X = opt.get("X")[opt.get("CV")[:, 0] <= 0]

            print(eps, len(X))

            if len(X) >= n_samples:
                break

            eps = eps * 0.9

        if len(X) > n_samples:
            I = select_points_with_maximum_distance(X, n_samples)
            X = X[I]

        return X



import warnings

import numpy as np
from cma.fitness_models import LQModel, SurrogatePopulationSettings
from cma.logger import LoggerDummy

from pymoo.algorithms.soo.nonconvex.cmaes import SimpleCMAES
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.optimize import minimize
from pymoo.problems.single import Sphere


class lqCMAES(SimpleCMAES):

    def __init__(self, **kwargs):
        super().__init__(advance_after_initialization=True, **kwargs)
        self.surrogate = SurrogatePopulation(None)

    def _infill(self):
        pass

    def _advance(self, **kwargs):
        X = super()._infill().get("X")
        X = self.norm.forward(X)

        surr = self.surrogate(X)

        f = None
        infills = []
        while True:
            try:
                x = surr.send(f)
                x = self.norm.backward(x)

                ind = Individual(X=x)
                self.evaluator.eval(self.problem, ind)
                infills.append(ind)
                f = ind.F[0]
            except StopIteration as ex:
                F = ex.value
                break

        self.es.tell(X, F)
        self.pop = Population.create(*infills)

        if self.es.stop():
            self.termination.force_termination = True


class SurrogatePopulation(object):

    def __init__(self,
                 fitness,
                 model=None,
                 model_max_size_factor=None,
                 tau_truth_threshold=None,
                 ):

        self.fitness = fitness
        self.model = model if model else LQModel()
        # set 2 parameters of settings from locals() which are not attributes of self
        self.settings = SurrogatePopulationSettings(locals(), 2, self)
        self.count = 0
        self.evaluations = 0
        self.logger = LoggerDummy(self, labels=['tau0', 'tau1', 'evaluated ratio'])
        self.logger_eigenvalues = LoggerDummy(self.model, ['eigenvalues'])

    class EvaluationManager:

        def __init__(self, X):
            """all is based on the population (list of solutions) `X`"""
            self.X = X
            self.evaluated = len(X) * [False]
            self.fvalues = len(X) * [np.nan]

        def add_eval(self, i, fval):
            """add fitness(self.X[i]), not in use"""
            self.fvalues[i] = fval
            self.evaluated[i] = True

        def eval(self, i, fitness, model_add):
            """add fitness(self.X[i]) to model data, mainly for internal use"""
            if self.evaluated[i]:  # need to decide what to do in this case
                raise ValueError("i=%d, evals=%d, len=%d" % (i, self.evaluations, len(self)))
            self.fvalues[i] = yield self.X[i]

            self.evaluated[i] = True
            model_add(self.X[i], self.fvalues[i])

        def eval_sequence(self, number, fitness, model_add, idx=None):
            """evaluate unevaluated entries of X[idx] until `number` entries are
            evaluated *overall*.

            Assumes that ``sorted(idx) == list(range(len(self.X)))``.

            ``idx`` defines the evaluation sequence.

            The post condition is ``self.evaluations == min(number, len(self.X))``.
            """
            if idx is None:
                idx = range(len(self.X))
            assert len(self.evaluated) == len(self.X)
            if not self.evaluations < number:
                warnings.warn("Expected evaluations=%d < number=%d, popsize=%d"
                              % (self.evaluations, number, len(self.X)))
            self.last_evaluations = number - self.evaluations  # used in surrogate loop
            for i in idx:
                if self.evaluations >= number:
                    break
                if not self.evaluated[i]:
                    yield from self.eval(i, fitness, model_add)
            else:
                if self.evaluations < number and self.evaluations < len(self.X):
                    warnings.warn("After eval: evaluations=%d < number=%d, popsize=%d"
                                  % (self.evaluations, number, len(self.X)))
                return
            assert self.evaluations == number or self.evaluations == len(self.X) < number

        def surrogate_values(self, model_eval, true_values_if_all_available=True,
                             f_offset=None):
            """return surrogate values of `model_eval` with smart offset.
            """
            if true_values_if_all_available and self.evaluations == len(self.X):
                return self.fvalues
            F_model = [model_eval(x) for x in self.X]
            if f_offset is None:
                f_offset = np.nanmin(self.fvalues)  # must be added last to prevent numerical erasion
            if np.isfinite(f_offset):
                m_offset = np.nanmin(F_model)  # must be subtracted first to get close to zero
                return [f - m_offset + f_offset for f in F_model]
            else:
                return F_model

        def __len__(self):
            """should replace ``len(self.X)`` etc, not fully in use yet"""
            return len(self.fvalues)

        @property
        def evaluation_fraction(self):
            return self.evaluations / len(self.fvalues)

        @property
        def evaluations(self):
            return sum(self.evaluated)

        @property
        def remaining(self):
            """number of not yet evaluated solutions"""
            return len(self.X) - sum(self.evaluated)

    def __call__(self, X):
        self.count += 1
        model = self.model  # convenience shortcut
        # a trick to see whether the population size has increased (from a restart)
        # model.max_absolute_size is by default initialized with zero
        # TODO: remove dependency on the initial value of model.max_absolute_size
        if self.settings.model_max_size_factor * len(X) > model.settings.max_absolute_size:
            if model.settings.max_absolute_size:  # do not reset in first call, in case model was initialized meaningfully
                model.reset()  # reset, because the population size changed
            model.settings.max_absolute_size = self.settings.model_max_size_factor * len(X)
        evals = SurrogatePopulation.EvaluationManager(X)
        self.evals = evals  # only for the record

        if 11 < 3:
            # make minimum_model_size unconditional evals in the first call and quit
            if model.size < self.settings.minimum_model_size:
                evals.eval_sequence(self.settings.minimum_model_size - model.size,
                                    self.fitness, model.add_data_row)
                self.evaluations += evals.evaluations
                model.sort(self.settings.model_sort_globally or evals.evaluations)
                return evals.surrogate_values(model.eval, self.settings.return_true_fitnesses)

        if 11 < 3 and self.count % (1 + self.settings.crazy_sloppy):
            return evals.surrogate_values(model.eval, f_offset=model.F[0])

        number_evaluated = int(1 + max((len(X) * self.settings.min_evals_percent / 100,
                                        3 / model.settings.truncation_ratio - model.size)))
        while evals.remaining:
            idx = np.argsort([model.eval(x) for x in X]) if model.size > 1 else None
            yield from evals.eval_sequence(number_evaluated, self.fitness,
                                           model.add_data_row, idx)
            model.sort(
                number_evaluated)  # makes only a difference if elements of X are pushed out on later adds in evals
            tau = model.kendall(self.settings.n_for_tau(len(X), evals.evaluations))
            if evals.last_evaluations == number_evaluated:  # first call to evals.eval_sequence
                self.logger.add(tau)  # log first tau
            if tau >= self.settings.tau_truth_threshold - self.settings.tau_truth_threshold_correction * evals.evaluation_fraction:
                break
            number_evaluated += int(np.ceil(number_evaluated / 2))
            """multiply with 1.5 and take ceil
                [1, 2, 3, 5, 8, 12, 18, 27, 41, 62, 93, 140, 210, 315, 473]
                +[1, 1, 2, 3, 4,  6,  9, 14, 21, 31, 47,  70, 105, 158]
            """

        model.sort(self.settings.model_sort_globally or evals.evaluations)
        model.adapt_max_relative_size(tau)
        self.logger.add(tau)  # log last tau
        self.evaluations += evals.evaluations
        if self.evaluations == 0:  # can currently not happen
            # a hack to have some grasp on zero evaluations from outside
            self.evaluations = 1e-2  # hundred zero=iterations sum to one evaluation
        self.logger.add(evals.evaluations / len(X))
        self.logger.push()
        return evals.surrogate_values(model.eval, self.settings.return_true_fitnesses)



import numpy as np

try:
    from GPyOpt.methods import BayesianOptimization
except:
    raise Exception("gpyopt not found. Please execute: 'pip install gpyopt'")

from pymoo.core.algorithm import Algorithm
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.util.display import SingleObjectiveDisplay


class EGO(Algorithm):

    def __init__(self,
                 display=SingleObjectiveDisplay(),
                 **kwargs):
        super().__init__(display=display, **kwargs)
        self.domain, self.func = None, None

    def _setup(self, problem, **kwargs):

        domain = []
        for k in range(problem.n_var):
            e = {'name': f"var_{k}", 'type': 'continuous', 'domain': (problem.xl[k], problem.xu[k])}
            domain.append(e)

        def func(x):
            ind = Individual(X=x[0])
            self.evaluator.eval(self.problem, ind)
            self.pop = Population.merge(self.pop, Population.create(ind))
            return ind.F[0]

        self.obj = BayesianOptimization(f=func, domain=domain)

    def _initialize_advance(self, **kwargs):
        self._advance()

    def _advance(self, **kwargs):
        obj = self.obj

        try:
            obj._update_model(obj.normalization_type)
        except np.linalg.linalg.LinAlgError:
            pass

        obj.num_acquisitions += 1
        obj.suggested_sample = obj._compute_next_evaluations()
        obj.X = np.vstack((obj.X, obj.suggested_sample))
        obj.evaluate_objective()




from pymoo.core.population import Population


class Archive:

    def __init__(self, survival, max_size=100, trunc_size=None, problem=None) -> None:
        super().__init__()
        self.sols = Population()
        self.survival = survival
        self.max_size = max_size
        self.trunc_size = trunc_size if trunc_size is not None else max_size
        self.problem = problem

    def add(self, sols):

        sols = Population.merge(self.sols, sols)

        if len(sols) > self.max_size:
            sols = self.survival.do(self.problem, sols, n_survive=self.trunc_size)

        self.sols = sols




import matplotlib.pyplot as plt
import numpy as np
from ezmodel.core.factory import models_from_clazzes
from ezmodel.core.selection import ModelSelection
from ezmodel.models.kriging import Kriging
from ezmodel.util.partitioning.crossvalidation import CrossvalidationPartitioning

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.algorithms.soo.nonconvex.ga_niching import NicheGA
from pymoo.core.callback import Callback
from pymoo.core.population import Population
from pymoo.operators.sampling.lhs import LHS
from pymoo.optimize import minimize
from pymoo.util.display import Display
from pymoo.util.normalization import ZeroToOneNormalization
from pymoo.util.termination.default import SingleObjectiveDefaultTermination
from pysamoo.core.algorithm import SurrogateAssistedAlgorithm
from pysamoo.experimental.acquisition import EI, AcquisitionProblem


# ---------------------------------------------------------------------------------------------------------
# Display
# ---------------------------------------------------------------------------------------------------------


class EGODisplay(Display):

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        self.output.append("f_opt", algorithm.opt[0].F[0])

        if algorithm.acq is None:
            self.output.append("f_new", "-")
            self.output.append("acq", "-")

            if algorithm.model_selection:
                self.output.append("regr", "-")
                self.output.append("corr", "-")
                self.output.append("ARD", "-")
        else:
            self.output.append("f_new", algorithm.infills.get("F").min())
            self.output.append("acq", - algorithm.infills[0].get("acq"))

            if algorithm.model_selection:
                gp = algorithm.surrogate
                self.output.append("regr", gp.regr)
                self.output.append("corr", gp.corr)
                self.output.append("ARD", gp.ARD)


# ---------------------------------------------------------------------------------------------------------
# Implementation
# ---------------------------------------------------------------------------------------------------------


class BayesianOptimization(SurrogateAssistedAlgorithm):

    def __init__(self,
                 acq_func=EI(),
                 model_selection=False,
                 adaptive_fmin=True,
                 display=EGODisplay(),
                 **kwargs):

        super().__init__(display=display, **kwargs)
        self.default_termination = SingleObjectiveDefaultTermination()

        self.model_selection = model_selection
        self.acq_func = acq_func
        self.adaptive_fmin = adaptive_fmin
        self.acq = None

    def _infill(self):

        # get all the points that have been evaluated yet
        X, F = self.archive.get("X", "F")

        # get he problem and the boundaries
        problem = self.problem
        xl, xu = problem.bounds()

        # the defaults for surrogate modeling - normalize the values to be between zero and one
        defaults = dict(norm_X=ZeroToOneNormalization(xl, xu))

        # rather the best model should be selected or simply the default kriging implementation taken
        if self.model_selection:
            models = models_from_clazzes(Kriging, **defaults)
            partitions = CrossvalidationPartitioning(k_folds=5, seed=1).do(X)
            model = ModelSelection(models).do(X, F[:, 0], partitions)
        else:
            model = Kriging(regr="linear", corr="gauss", ARD=True, **defaults)
            model.fit(X, F[:, 0])

        if self.adaptive_fmin:

            # get the acquisition problem to be optimized
            acq = robust_fmin_acquisition(problem, model, self.acq_func, self.archive)

        else:
            # just use the minimum (even though this can lead to precision issues)
            _min = F[:, 0].argmin()
            f_min = model.predict(X[_min])[0, 0]
            acq = AcquisitionProblem(problem, model, self.acq_func, f_min=f_min)

        # use a bigger latin hypercube in the beginning because the problem might be highly multi-modal
        sampling = LHS().do(problem, 500)

        algorithm = NicheGA(pop_size=50, sampling=sampling)

        termination = SingleObjectiveDefaultTermination(nth_gen=1, n_last=5)

        res = minimize(acq,
                       algorithm,
                       termination,
                       verbose=False
                       )

        X, F = res.opt.get("X", "F")

        self.acq = acq
        self.surrogate = model
        return Population.new(X=X, acq=F[0])

    def _set_optimum(self):
        self.opt = FitnessSurvival().do(self.problem, self.archive, n_survive=1)


def robust_fmin_acquisition(problem, model, acq_func, points):
    X, F = points.get("X", "F")

    sorted_by_pred = np.sort(model.predict(X)[:, 0])

    n_intervals = min(20, len(X))
    n_points = 500

    interval = int(len(sorted_by_pred) / n_intervals)

    for cnt in range(n_intervals):

        # increase the index for f_min in each iteration
        f_min = sorted_by_pred[cnt * interval]

        # create the acquisition problem and radomly sample
        acq = AcquisitionProblem(problem, model, acq_func, f_min=f_min)
        sampling = LHS().do(problem, n_points)

        # the maximum value found of during random sampling
        max_prob_imprv = (- acq.evaluate(sampling.get("X"))).max()

        # print(cnt, sorted_by_pred[0], f_min, max_prob_imprv)

        # if the value is not very small then the fmin is okay to be used
        if max_prob_imprv > 1e-2:
            break

    return acq


class EGOVisualization(Callback):

    def notify(self, algorithm):
        problem = algorithm.problem
        if problem.n_var > 1 or problem.n_obj > 1 or algorithm.surrogate is None:
            return

        fig = plt.figure()

        gs = fig.add_gridspec(4, 1)
        plt_func = fig.add_subplot(gs[:3])
        plt_acq = fig.add_subplot(gs[3])

        X = algorithm.pop.get("X")
        F = problem.evaluate(X)
        infill = algorithm.infills[0]
        plt_func.scatter(X, F, color="red")
        acq = algorithm.acq

        mesh = np.linspace(problem.xl[0], problem.xu[0], 1000)[:, None]

        gp = algorithm.surrogate
        mu, sigma = gp.predict(mesh, return_values_of=["y", "sigma"])

        plt_func.fill_between(mesh[:, 0], (mu - 2 * sigma)[:, 0], (mu + 2 * sigma)[:, 0], alpha=0.2, color='k')

        plt_func.scatter(infill.X, infill.F, color="red", s=100, marker="x")
        plt_func.plot(mesh, gp.predict(mesh), color="red")

        plt_func.axvline(x=algorithm.infills[0].X, color="black", linestyle='dashed')

        plt_func.plot(mesh, problem.evaluate(mesh), color="black")

        plt_acq.plot(mesh, acq.evaluate(mesh), color="blue")
        plt_acq.scatter(infill.X, acq.evaluate(infill.X), color="red", s=100, marker="x")

        plt.show()



import numpy as np

from pymoo.core.sampling import Sampling
from pymoo.operators.sampling.lhs import LHS
from pymoo.util.misc import cdist
from pymoo.util.normalization import normalize


def select_points_with_maximum_distance(X, n_select, selected=[]):
    n_points, n_dim = X.shape

    # calculate the distance matrix
    D = cdist(X, X)

    # if no selection provided pick randomly in the beginning
    if len(selected) == 0:
        selected = [np.random.randint(len(X))]

    # create variables to store what selected and what not
    not_selected = [i for i in range(n_points) if i not in selected]

    # remove unnecessary points
    dist_to_closest_selected = D[:, selected].min(axis=1)

    # now select the points until sufficient ones are found
    while len(selected) < n_select:
        # find point that has the maximum distance to all others
        index_in_not_selected = dist_to_closest_selected[not_selected].argmax()
        I = not_selected[index_in_not_selected]

        # add the closest distance to selected point
        is_closer = D[I] < dist_to_closest_selected
        dist_to_closest_selected[is_closer] = D[I][is_closer]

        # add it to the selected and remove from not selected
        selected.append(int(I))
        not_selected = np.delete(not_selected, index_in_not_selected)

    return selected


class CustomLHS(LHS):

    def __init__(self, iterations=100, others=None, **kwargs) -> None:
        super().__init__(iterations=iterations, **kwargs)
        self.others = others
        self.norm_others = None

    def _do(self, problem, n_samples, **kwargs):

        if self.others is not None:
            xl, xu = problem.bounds()
            self.norm_others = normalize(self.others, xl, xu)

        return super()._do(problem, n_samples, **kwargs)

    def _calc_score(self, X):
        val = super()._calc_score(X)

        if self.norm_others is not None and len(self.norm_others) > 0:
            D = cdist(X, self.norm_others)
            val = min(val, np.min(D))

        return val


class RejectionConstrainedSampling(Sampling):

    def __init__(self,
                 func_eval_constr,
                 batch_size=None,
                 n_multiplier=2,
                 max_iter=100
                 ):
        super().__init__()
        self.max_iter = max_iter
        self.n_multiplier = n_multiplier
        self.batch_size = batch_size
        self.func_eval_constr = func_eval_constr

    def _do(self, problem, n_samples, **kwargs):

        n_points = self.batch_size
        if n_points is None:
            n_points = 2 * n_samples

        ret = np.zeros((0, problem.n_var))

        for k in range(self.max_iter):

            if len(ret) >= self.n_multiplier * n_samples:
                break

            else:

                sampling = CustomLHS(others=ret)

                X = sampling.do(problem, n_points).get("X")

                CV = self.func_eval_constr(X)
                is_feasible = (CV <= 0)
                X = X[is_feasible]

                ret = np.row_stack([ret, X])

        if len(ret) > n_samples:
            I = select_points_with_maximum_distance(ret, n_samples)
            ret = ret[I]

        return ret


import setuptools

from pysamoo.version import __version__

# ---------------------------------------------------------------------------------------------------------
# GENERAL
# ---------------------------------------------------------------------------------------------------------


__name__ = "pysamoo"
__author__ = "Julian Blank"
__url__ = "https://anyoptimization.com/projects/pysamoo/"

data = dict(
    name=__name__,
    version=__version__,
    author=__author__,
    url=__url__,
    python_requires='>=3.7',
    author_email="blankjul@msu.edu",
    description="Surrogate-Assisted Multi-objective Optimization",
    license='GNU AFFERO GENERAL PUBLIC LICENSE (AGPL)',
    keywords="surrogate, metamodel, bayesian optimization",
    install_requires=["pymoo>0.5.0", "ezmodel"],
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics'
    ]
)


# ---------------------------------------------------------------------------------------------------------
# METADATA
# ---------------------------------------------------------------------------------------------------------


# update the readme.rst to be part of setup
def readme():
    with open('README.rst') as f:
        return f.read()


def packages():
    return ["pysamoo"] + ["pysamoo." + e for e in setuptools.find_packages(where='pysamoo')]


data['long_description'] = readme()
data['packages'] = packages()

setuptools.setup(**data)



from copy import deepcopy

import numpy as np
from ezmodel.core.benchmark import Benchmark
from ezmodel.core.partitioning import merge_and_partition
from ezmodel.util.partitioning.crossvalidation import CrossvalidationPartitioning
from pymoo.core.population import Population
from pymoo.util.misc import from_dict
from pymoo.util.sliding_window import SlidingWindow

from pysamoo.core.indicator import INDICATORS


class Target:

    def __init__(self,
                 label,
                 models,
                 n_folds=5,
                 n_max_performances=5,
                 n_max_benchmarks=0,
                 indicators=INDICATORS):
        """

        Parameters
        ----------
        label : tuple
            This describes the type and what role it plays in the problem later on. For instance, if this target
            models the first constraint ('G', 0) would be the corresponding type.

        models : list
            A list of all models which are kept track of for this target.

        n_folds : int
            The number of folds if cross-validation is applied (happens only the first time)

        n_max_performances : int
            The number of maximum last performances to make a decision

        n_max_benchmarks : int
            The maximum number of benchmarks (only to review later)

        indicators : list
            A list of tuples (name, sign, func) defining what indicators should be calculated after
            each of the benchmarks.

        """

        self.label = label
        self.models = models
        self.n_folds = n_folds

        # the indicators to be calculated for this target
        self.indicators = indicators

        # this is the storage for the most recent benchmarks - full experiment
        self.benchmarks = SlidingWindow(size=n_max_benchmarks)

        # this keeps track of the past performances - each model gets one
        self.performances = {}
        for key in models.keys():
            self.performances[key] = SlidingWindow(size=n_max_performances)

        # the current name of the model which is used
        self.model = None

        # the actual model fitting the data points provided
        self.obj = None

        # the best model set by the last validation
        self.best = None

    def validate(self, trn, tst=None, find_best=True, **kwargs):

        # get the values to be predicted
        X, y = trn.get("X"), self._get_y(trn)

        if tst is None:
            X, y, partitions = X, y, CrossvalidationPartitioning(self.n_folds).do(len(trn))
        else:
            _X, _y = tst.get("X"), self._get_y(tst)
            X, y, partitions = merge_and_partition((X, y), (_X, _y))

        # do the benchmark for the specific target given the partitions
        obj = Benchmark(self.models, raise_exception=False).do(X, y, partitions=partitions)
        benchmark = obj.results(only_successful=False, as_list=False, include_metadata=True)
        self.benchmarks.append(benchmark)

        for model in self.models.keys():

            results = benchmark["results"][model]

            # for each of the run, execute all the performance indicators
            for run in results["runs"]:
                ret = self._indicators(benchmark, run)
                self.performances[model].append(ret)

        # if a list of indicators for selecting the best model is provided, do it
        if find_best:
            self.best = self.find_best(**kwargs)

    def find_best(self, indicator=["kendall_tau", "mae"], exclude=[]):

        models = [model for model in self.models.keys() if model not in exclude]
        perf = self.performances

        # filter the models if one run as failed during benchmark
        models = np.array([m for m in models if None not in perf[m]])

        assert len(models) > 0, "Fitting each of the models has failed at least once in the benchmark."

        for entry in indicator:

            # get the performances from the the n_max_performance iterations
            v = np.array([self.performance(entry, model=model) for model in models])

            # multiply by the sign to consider minimization and maximization
            v *= self.indicators[entry]["sign"]

            # find the best model or models - if there is a tie
            models = models[v == np.min(v)]

            if len(models) == 1:
                break

        # finally find the best model using the indicator
        return np.random.choice(models)

    def fit(self, sols):
        assert self.best is not None, "You need to do one initial validation to find the best model for this target."

        # finally fit the model with the data
        obj = deepcopy(self.models[self.best])

        # get in and output
        X, y = sols.get("X"), self._get_y(sols)
        obj.fit(X, y)

        self.model = self.best
        self.obj = obj

    def performance(self, indicator, model=None, func=np.mean):
        if model is None:
            model = self.best

        assert model is not None, "Either provide a model or set one through validation."

        v = np.array([e[indicator] for e in self.performances[model]])

        if func:
            v = func(v)

        return v

    def predict(self, X, out):
        assert self.obj is not None, "The target has not been fitted yet."
        v = self.obj.predict(X)
        key, index = self.label
        out.get(key)[:, [index]] = v

    def _indicators(self, benchmark, run):

        X, y, partitions = from_dict(benchmark, "X", "y", "partitions")

        if run["success"]:
            ret = {}

            partition = run["partition"]
            trn, tst = partitions[partition]

            y_true = y[tst]
            y_hat = run["y_hat"]

            for name, v in self.indicators.items():
                ret[name] = v["func"](y_true, y_hat, trn_y=y[trn])

            return ret

    def _get_y(self, sols):
        """
        Parameters
        ----------
        sols : Population
            A set of solutions.

        Returns
        -------
        Y : np.array
            The Y values to be modeled by the surrogate

        """
        key, index = self.label
        return sols.get(key)[:, index]

    def __repr__(self) -> str:
        return f"({self.label[0]}, {self.label[1]}) : {super(Target, self).__repr__()}"
